"""recompute.py

A sweet tool for Remote Execution.
This is the high-level user interface.
The user is exposed to the program through this module.
We bombard the user with a suite of features that enable comfortable execution of code in remote machines.
Each feature, named `mode` here, corresponds to a particular operation.
Scroll down for a table of available commands, options and how to use them.

"""
from recompute import utils
from recompute.instance import Instance
from recompute.config import ConfigManager
from recompute.instance import InstanceManager
from recompute.remote import Remote
from recompute.bundle import Bundle
from recompute.remote import VOID_CACHE

from getpass import getpass

import argparse
import logging
import os

# setup logger
logger = logging.getLogger(__name__)

# man page
MAN_DOCU = """
                         _ __   ___ 
                        | '__| / _ \\
                        | |   |  __/
                        |_|    \___|

+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| Mode     | Description                                         | Options               | Example                             |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| init     | Setup current directory for remote execution        | --instance-idx        | $re init                            |
|          |                                                     |                       | $re init --instance-idx=1           |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| rsync    | Use rsync to synchronize local files with remote    | --force               | $re rsync                           |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| sshadd   | Add a new instance to config                        | --instance            | $re sshadd --instance="usr@host"    |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| install  | Install pypi packages in requirements.txt in remote | cmd, --force          | $re install                         |
|          |                                                     |                       | $re install "pytorch tqdm"          |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| sync     | Synchronous execution of "args.cmd" in remote       | cmd, --force, --rsync | $re sync "python3 x.py"             |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| async    | Asynchronous execution of "args.cmd" in remote      | cmd, --force, --rsync | $re async "python3 x.py"            |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| log      | Fetch log from remote machine                       | --loop, --filter      | $re log                             |
|          |                                                     |                       | $re log --loop=2                    |
|          |                                                     |                       | $re log --filter="pattern"          |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| list     | List out processes alive in remote machine          | --force               | $re list                            |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| kill     | Kill a process by index                             | --idx                 | $re kill                            |
|          |                                                     |                       | $re kill --idx=1                    |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| purge    | Kill all remote process that are alive              | None                  | $re purge                           |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| ssh      | Create an ssh session in remote machine             | None                  | $re ssh                             |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| notebook | Create jupyter notebook in remote machine           | --run-async           | $re notebook                        |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| push     | Upload file to remote machine                       | cmd                   | $re push "x.py y/"                  |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| pull     | Download file from remote machine                   | cmd                   | $re pull "y/z.py ."                 |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| data     | Download data from web into data/ folder of remote  | cmd                   | $re data "url1 url2 url3"           |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
| man      | Show this man page                                  | None                  | $re man                             |
+----------+-----------------------------------------------------+-----------------------+-------------------------------------+
"""

# parse command-line arguments
parser = argparse.ArgumentParser(
    description='recompute.py -- A sweet tool for remote computation'
    )
# NOTE : ffs! write a descriptive help for `mode`
parser.add_argument('mode', type=str,
    help='(init/sync/async/rsync/install/log/list/kill/purgessh/notebook/conf/probe/data/pull/push/sshadd/man) recompute mode')
parser.add_argument('cmd', nargs='?', default='None',
    help='command to run in remote system')
parser.add_argument('--remote-home', nargs='?', default='projects/',
    help='remote projects/ directory')
parser.add_argument('--urls', nargs='?', default='',
    help='comma-separated list of URLs')
parser.add_argument('--instance', nargs='?', default='',
    help='[username@host] config.remotepass is used')
parser.add_argument('--filter', nargs='?', default='',
    help='keyword to filter log')
parser.add_argument('--loop', nargs='?', default='',
    help='number of seconds to wait to fetch log')
parser.add_argument('--idx', nargs='?', default='',
    help='process idx to operate on')
parser.add_argument('--name', nargs='?', default='runner',
    help='name of process')
parser.add_argument('--instance-idx', nargs='?', default=0,
    help='remote instance to use')
parser.add_argument('--force', default=False, action='store_true',
    help='clear cache')
parser.add_argument('--no-force', dest='force', action='store_false')
parser.add_argument('--run-async', default=False, action='store_true',
    help='Execute commands async')
parser.add_argument('--no-run-async', dest='run_async', action='store_false')
parser.add_argument('--rsync', default=True, action='store_true',
    help='Update files in remote machine')
parser.add_argument('--no-rsync', dest='rsync', action='store_false')
args = parser.parse_args()


def init():
  """Setup current directory for remote execution

  * Setup remote instance for execution
  * Create local configuration files
  * Bundle up local repository
  * Sync files
  * Install Dependencies

  Returns
  -------
  remote.Remote
    An instance of Remote class
  """
  # get configuration manager
  confman = ConfigManager()
  # build instance manager
  instanceman = InstanceManager(confman)
  # create default instance
  instance_idx = int(args.instance_idx) if args.instance_idx else None
  instance = instanceman.get(instance_idx)
  # create bundle
  bundle = Bundle()
  # create remote instance handle
  remote = Remote(instance, bundle, remote_home=args.remote_home)
  # sync files
  remote.rsync()
  # install from requirements.txt
  remote.install_deps()
  return remote


def cache_exists():
  """Does cache exist?"""
  return os.path.exists(VOID_CACHE)


def get_remote():
  """Get an instance of Remote from cache or create anew"""
  if cache_exists():
    logger.info('Cache exists')
    return Remote()

  logger.info('Cache doesn\'t exist')
  return init()


def main():  # package entry point

  """ Boilerplate """
  # get configuration manager
  confman = ConfigManager()
  # build instance manager
  instanceman = InstanceManager(confman)

  # ------------ man ------------- #
  if args.mode == 'man':
    """ Mode : Show man page """
    print(MAN_DOCU)
    exit()

  # ------------ conf ------------ #
  elif args.mode == 'conf':  # generate config file
    """ Mode : Generate configuration file """
    config = confman.generate(force=args.force)
    if not config:
      logger.info('config exists; use --force to overwrite it')
      print('config exists; use --force to overwrite it')

  # ------------ sshadd ---------- #
  elif args.mode == 'sshadd':  # add remote instance
    """ Mode : Add remote instance to config """
    try:
      assert args.instance  # make sure user@host is given as input
      # get password from user
      password = getpass('Password:')
      # . create Instance instance
      # .. parse user@host
      instance = Instance(password=password).resolve_str(args.instance)
      # add instance to config
      instanceman.add_instance(instance)
    except AssertionError:
      logger.error('Invalid/Empty instance')

  # ------------ probe ----------- #
  elif args.mode == 'probe':  # probe remote machines
    """ Mode : Probe remote machines """
    print(instanceman.probe(force=args.force))

  # ------------ data ------------ #
  elif args.mode == 'data':
    """ Mode : GET data from web """
    try:
      assert args.urls
      get_remote().download(args.urls.split(' '), run_async=args.run_async)
    except AssertionError:
      logger.error('Input a list of URLs to download')

  # ------------ init ------------ #
  elif args.mode == 'init':
    """ Mode : Initialize project in current directory """
    init()  # create remote anew

  # ------------ sync ------------ #
  elif args.mode == 'sync':
    """ Mode : Sync Execute command in remote machine """
    assert args.cmd  # user inputs command to exec in remote
    # create remote from cache
    remote = get_remote()
    # look for python execution
    if 'python' in args.cmd and args.rsync:  # TODO : this is pretty hacky;
      remote.rsync(update=args.force)          # you are better than this!
    # blocking execute `cmd` in remote
    remote.execute([args.cmd], log=True, name=args.name)

  # ------------ async ----------- #
  elif args.mode == 'async':
    """ Mode : Async Execute command in remote machine """
    assert args.cmd
    # get remote
    remote = get_remote()
    # look for python execution
    if 'python' in args.cmd and args.rsync:
      remote.rsync(update=args.force)
    # async execute `cmd` in remote
    remote.async_execute([args.cmd], name=args.name)

  # ------------ rsync ----------- #
  elif args.mode == 'rsync':
    """ Mode : Rsync files """
    # create remote from cache
    get_remote().rsync(update=args.force)

  # ------------ install --------- #
  elif args.mode == 'install':
    """ Mode : Rsync files """
    if not args.cmd:
      # create remote from cache
      get_remote().install_deps(update=args.force)
    else:
      # space separated pypi packages
      get_remote().install(args.cmd.split(' '))

  # ------------ log ------------- #
  elif args.mode == 'log':  # copy log from remote
    """ Mode : Copy log from remote machine """
    # get remote
    remote = get_remote()
    # delete local log
    # NOTE : i'm not sure if i should do this!
    if os.path.exists(remote.local_logfile):  # if it exists
      os.remove(remote.local_logfile)
    if args.loop:  # --------- loop ----------- #
      """ Mode : Copy log from remote machine in a loop """
      remote.loop_get_remote_log(int(args.loop), args.filter)
    else:  # --------------- no loop ---------- #
      log = remote.get_remote_log(args.filter)
      print(utils.parse_log(log))

  # ------------ list ------------ #
  elif args.mode == 'list':  # list of processes
    """ Mode : List remote processes """
    print(utils.tabulate_processes(
        get_remote().list_processes(force=True)
        ))

  # ------------ kill ------------ #
  elif args.mode == 'kill':  # kill process
    """ Mode : Interactive kill """
    remote = get_remote()
    # print table of processes
    print(utils.tabulate_processes(
      remote.list_processes(force=True)
      ))
    # resolve index of proc
    idx = int(args.idx) if args.idx else None
    if not args.idx:
      # get index from user
      try:
        idx = int(input('Process to kill (index) : '))
      except KeyboardInterrupt:
        exit()  # the user chikened out!
    # kill process
    remote.kill(idx)

  # ------------ purge ----------- #
  elif args.mode == 'purge':  # kill them all
    get_remote().kill(0)

  # ------------ ssh ------------- #
  elif args.mode == 'ssh':  # start an ssh session
    """ Mode : Create an ssh session """
    get_remote().get_session()

  # ------------ notebook -------- #
  elif args.mode == 'notebook':  # start an ssh session
    """ Mode : Create and connect to remote notebook server """
    get_remote().start_notebook(args.run_async)

  # ------------ pull ------------ #
  elif args.mode == 'pull':  # copy log from remote
    """ Mode : Download file from remote machine """
    assert args.cmd  # make sure files are provided for download
    # NOTE : relative path is used
    # NOTE : copy one file/directory at a time
    remote = get_remote()

    # parse list of files/path
    filepaths = args.cmd.split(' ')

    try:  # make sure only one file is copied at a time
      assert len(filepaths) <= 2
    except AssertionError:
      logger.error('Copy one file/directory at a time')
      exit()

    # get absolute paths of local file and remote path
    remotefile = os.path.join(remote.remote_dir, filepaths[0])
    localpath = None if len(filepaths) == 1 \
        else os.path.join(remote.bundle.path, filepaths[1])

    remote.get_file_from_remote(remotefile, localpath)

  # ------------ push ------------ #
  elif args.mode == 'push':  # copy log from remote
    """ Mode : Upload file to remote machine """
    assert args.cmd  # make sure files are provided for download
    # NOTE : relative path is used
    # NOTE : copy one file/directory at a time
    remote = get_remote()
    # parse list of files/path
    filepaths = args.cmd.split(' ')

    try:  # make sure only one file is copied at a time
      assert len(filepaths) <= 2
    except AssertionError:
      logger.error('Copy one file/directory at a time')
      exit()

    # get absolute paths of local file and remote path
    localfile = os.path.join(remote.bundle.path, filepaths[0])
    remotepath = None if len(filepaths) == 1 \
        else os.path.join(remote.remote_dir, filepaths[1])

    # copy file to remote
    remote.copy_file_to_remote(localfile, remotepath)

  else:
    logger.error('Something went wrong! Check command-line arguments')



"""process.py

A suite of functions to execute commands in local and remote machines,
and to keep track (manage) of created processes.

"""
import os
import subprocess
import logging
import signal

from recompute import cmd

# setup logger
logger = logging.getLogger(__name__)


def is_process_alive(pid):
  """Check if a process corresponding to `pid` is alive

  Parameters
  ----------
  pid : int
    Process id

  Returns
  -------
    bool
      `True` if the process is alive, `False` otherwise
  """
  try:
    return os.kill(pid, 0) is None
  except ProcessLookupError:
    return


def is_remote_process_alive(pid, instance):
  """Check if a process is alive in remote device

  Parameters
  ----------
  pid : int
    Process id
  instance : instance.Instance
    an Instance object (remote device) to run process on

  Returns
  -------
    bool
      `True` if the process is alive in remote device
      `False` otherwise
  """
  _, output = remote_execute(cmd.PROCESS_PID_LINUX.format(pid=pid), instance)
  return str(pid) in output


def kill_process(pids):
  """Kill processes based on their pids

  Parameters
  ----------
  pids : list
    A list of process id's to kill
  """
  for pid in pids:
    os.kill(pid, signal.SIGTERM)


def kill_remote_process(pids, instance):
  """Kill processes in remote device (`instance`)

  Parameters
  ----------
  pids : list
    A list of process id's to kill
  instance : instance.Instance
    Instance object corresponding to remote device

  Returns
  -------
  str
    Result of `kill` command
  """
  _, output = remote_execute(cmd.kill_procs(pids), instance)
  return output


def fetch_stderr(cmdstr):
  """Fetch STDERR from executing `cmdstr`

  Parameters
  ----------
  cmdstr : str
    Command to be executed

  Returns
  -------
  str
    STDERR of executed command
  """
  # create process
  process = subprocess.Popen([cmdstr, '...'],
      stdout=subprocess.PIPE,
      stderr=subprocess.PIPE,
      shell=True)
  stdout, stderr = process.communicate()
  logger.info(cmdstr)
  # read from stderr
  logger.info('ERR : {}'.format(stderr.decode('utf-8')))
  return stderr.decode('utf-8')


def execute(cmdstr, run_async=False):
  """Execute `cmdstr` and return results

  Parameters
  ----------
  cmdstr : str
    Command to be executed
  run_async : bool, optional
    When set to `True` the command is executed asynchronously
    When set to `False`, blocking execution happens (default False)

  Returns
  -------
  pid : int
    Process id of command executed
  output : str
    STDOUT of execution as a string
    `None` is returned when executed asynchronously
  """
  # set stdout PIPE
  stdout = open(os.devnull) if run_async else subprocess.PIPE
  # create process
  process = subprocess.Popen([cmdstr, '...'], stdout=stdout, shell=True)
  logger.info(cmdstr)

  if run_async:  # return PID if running async
    return process.pid, None

  try:  # else wait for process to complete
    # get stdout
    output_bytes, error = process.communicate()
    output_str = output_bytes.decode('utf-8')
    logger.info(output_str)
    return process.pid, output_str
  except KeyboardInterrupt:
    logger.error('Keyboard Interrupt')
    return None, None
  except Exception:
    logger.error('Execution Failed!')
    return None, None


def async_execute(cmdstr):
  """Execute `cmdstr` asynchronously

  Parameters
  ----------
  cmdstr : str
    Command to be executed

  Returns
  -------
  pid : int
    Process id of command executed
  output : NoneType
    None
  """
  return execute(cmdstr, run_async=True)


def remote_execute(cmdstr, instance, bypass_subprocess=False):
  """Execute `cmdstr` in remote device given by `instance`

  Parameters
  ----------
  cmdstr : str
    Command to be executed
  instance : instance.Instance
    Instance of remote device
  bypass_subprocess : bool, optional
    When set to `True`, `os.system` is used for execution, (None, None) is returned
    When set to `False`, subprocess module is used for execution (default False)

  Returns
  -------
  tuple
    (pid, output) Process id and STDOUT of execution
  """
  _header = cmd.SSH_HEADER.format(password=instance.password)
  _body = cmd.SSH_EXEC.format(
      username=instance.username,
      host=instance.host, cmd=cmdstr
      )
  if bypass_subprocess:
    _body = cmd.SSH_EXEC_PSEUDO_TERMINAL.format(
          username=instance.username,
          host=instance.host, cmd=cmdstr
          )
    os.system(' '.join([_header, _body]))
    return None, None

  return execute(' '.join([_header, _body]))


def remote_async_execute(cmdstr, instance, logfile='/dev/null'):
  """Execute `cmdstr` in remote device given by `instance`

  Parameters
  ----------
  cmdstr : str
    Command to be executed
  instance : instance.Instance
    Instance of remote device
  bypass_subprocess : bool
    When set to `True`, `os.system` is used for execution, (None, None) is returned
    When set to `False`, subprocess module is used for execution

  Returns
  -------
  tuple
    (pid, output) `pid` contains the process id of command executed
    `output` is always `None` for aysnc execution
  """
  _header = cmd.SSH_HEADER.format(password=instance.password)
  _body = cmd.SSH_EXEC_ASYNC.format(
      username=instance.username,
      host=instance.host, cmd=cmdstr,
      logfile=logfile
      )
  _, output = execute(' '.join([_header, _body]))
  # parse output to get PID of remote process
  pid = int(output.replace('\n', '').strip())
  return pid, None


def create_runner(path, commands, logfile, run_async=False, name='re.runner'):
  """Create a bash script for executing `commands` sequentially in remote system

  Parameters
  ----------
  path : str
    Path in remote device, from where `commands` should be executed
  commands : list
    A list of commands to be executed
  logfile : str
    A file where the output of execution should be redirected
  run_async : bool, optional
    When set to `True` the script is executed asynchronously
    When set to `False`, blocking execution happens (default False)
  name : str, optional
    Name of the script which contains `commands`
    The script that will be executed

  Returns
  -------
  str
    Name of the script
  """
  # . set traps
  # .. change to path
  lines = cmd.make_traps() + [ cmd.CD.format(path=path) ]
  # async execution
  for i, command in enumerate(commands):
    if run_async:  # redirect stdout/stderr to log file
      command = cmd.REDIRECT_STDOUT.format(command=command, logfile=logfile)
      if i < len(commands) - 1:
        command = '{} &'.format(command)  # push to background
      else:  # add EOF to log if last command
        command = '({} && echo EOF >> {}) &'.format(command, logfile)
    # add to list of lines
    lines.append(command)
  # end with wait if "run_async"
  lines = lines if not run_async else lines + [ cmd.WAIT ]
  # lines = lines + [ cmd.WAIT ]
  # write to disk
  logger.info('Write to file')
  with open(name, 'w') as f:
    for line in lines:
      logger.info(line)
      f.write(line)
      f.write('\n')

  return name



"""instance.py

We model (`username`, `host`, `password`) as an "instance".
`Instance` class holds username, host and password information.
It can be saved to and read from the global configuration file.
`InstanceManager` class reads instances from the configuration file by interacting with `ConfigManager`.

"""
from recompute import process
from recompute import cmd
from recompute import utils

import logging

import pickle
import os

# setup logger
logger = utils.get_logger(__name__)
# table cache
PROBE_CACHE = '.recompute/table'


class Instance(object):
  """Instance is a container for (`username`, `password`, `host`)."""

  def __init__(self, username=None, password=None, host=None):
    """
    Parameters
    ----------
    username : str, optional
      User Name to log in to remote machine (default None)
    password : str, optional
      Password to log in to remote machine (default None)
    host : str, optional
      IP address or host name of remote machine (default None)
    """
    self.username = username
    self.password = password
    self.host = host

  def resolve_str(self, loginstr):
    """Create Instance object from string of type "username@host"

    Parameters
    ----------
    loginstr : str
      String of type "username@host"

    Returns
    -------
      instance.Instance
      An Instance object build based on `loginstr`
    """
    try:
      self.username, self.host = loginstr.strip().split('@')
      return self
    except ValueError:
      logging.error('Check Login [{}]'.format(loginstr))
      exit()

  def resolve_conf(self, conf):
    """Create Instance object from string of type "username@host"

    Parameters
    ----------
    conf : configparser.SectionProxy
      An instance section from config file

    Returns
    -------
      instance.Instance
      An Instance object build based on `conf`
    """
    self.username = conf['username']
    self.password = conf['password']
    self.host = conf['host']
    return self

  def __repr__(self):
    """Representation of form "username@host"

    Returns
    -------
    str
      "username@host" representation
    """
    return '{username}@{host}'.format(
        username=self.username, host=self.host
        )

  def __eq__(self, other):
    """Equivalence check of Instance objects

    Returns
    -------
    bool
      `True` if two instances are equivalent, `False` otherwise
    """
    return self.username == other.username and \
        self.password == other.password and \
        self.host == other.host


class InstanceManager(object):
  """InstanceManager manages instances by interacting with ConfigManager."""

  def __init__(self, confman):
    """
    Parameters
    ----------
    confman : config.ConfigManager
      Configuration Manager object
    """
    self.confman = confman

  def add_instance(self, instance):
    """Add an instance to global config

    Parameters
    ----------
    instance : instance.Instance
      An Instance object
    """
    # make sure the instance is active
    assert self.is_active(instance), 'Instance Inactive'
    # check if it's a duplicate
    assert len([ i for i in self.get_all()
      if i == instance ]) == 0, 'Duplicate Instance'
    # add instance to config file
    self.confman.add_instance(instance)

  def is_active(self, instance):
    """Is an instance active?

    Parameters
    ----------
    instance : instance.Instance
      An Instance object

    Returns
    -------
    bool
      `True` if instance is active, `False` otherwise
    """
    return not process.fetch_stderr( ' '.join([
      cmd.SSH_HEADER.format(password=instance.password),
      cmd.SSH_TEST.format(username=instance.username, host=instance.host)
      ]))

  def get(self, idx=None):
    """Find instance section from config file.

    Create an Instance object from the read instance section.

    Parameters
    ----------
    idx : int, optional
      Index of instance section in config file (default None)

    Returns
    -------
    instance.Instance
      An Instance object build on instance section read from config
    """
    # get instance config
    instance = self.confman.get_instance(idx)
    # make sure the instance exists in config
    assert instance, 'Instance inactive or absent in config'
    # return an instance
    return Instance(instance['username'], instance['password'], instance['host'])

  def get_all(self):
    """Return a list of instances from config file.

    Read all the instance sections in config file.
    Create a list of Instance objects from the read sections.

    Returns
    -------
    list
      A list of Instance objects read from config
    """
    return [ Instance().resolve_conf(instance)
        for instance in self.confman.get_instances() ]

  def get_active(self):
    """Return a list of active instances.

    Create a list of Instance objects from the read sections.
    Filter out the inactive instances.

    Returns
    -------
    list
      A list of active Instance objects read from config
    """
    return [ instance for instance in self.get_all()
        if self.is_active(instance) ]

  def fetch(self):
    """Fetch an active instance by reading config file

    Returns
    -------
    instance.Instance
      An active Instance object
    """
    # get all instances
    for instance in self.get_all():
      if self.is_active(instance):  # find an instance that's active
        return instance

  def probe(self, force=False):
    """Probe all the active instances for the following information.

    * Free GPU Memory
    * Free Disk Space

    Parameters
    ----------
    force : bool, optional
      When set to `True`, probes remote devices for information
      When `False`, read from local cache (default False)

    Returns
    -------
    prettytable.PrettyTable
      A pretty-looking table of required information
    """
    if not force and os.path.exists(PROBE_CACHE):
      return utils.tabulate_instances(pickle.load(open(PROBE_CACHE, 'rb')))

    # init dictionary of instances
    instances = {}
    # get active instances
    for instance in self.get_active():
      # init row
      instances[str(instance)] = [ str(instance), 'active', '-', '-' ]
      logger.info(instance)
      try:
        # gather info from remote machines
        # TODO : execute them in one session
        free_gpu_memory = int(
            process.remote_execute(cmd.GPU_FREE_MEMORY, instance)[-1]
            )
        logger.info('FREE GPU')
        logger.info(free_gpu_memory)
        free_disk_space = utils.parse_free_results(
            process.remote_execute(cmd.DISK_FREE_MEMORY, instance)[-1]
            )
        logger.info('FREE DISK')
        logger.info(free_disk_space)
        # update dictionary
        instances[str(instance)][2] = free_gpu_memory
        instances[str(instance)][3] = free_disk_space
      except ValueError:
        # something wrong? -> blame the host..
        instances[str(instance)][1] = 'active'
        continue
    # cache table
    pickle.dump(instances, open(PROBE_CACHE, 'wb'))
    return utils.tabulate_instances(instances)



"""config.py

The Configuration Manager is an interface to interact with the global configuration file.
It can generate, load and update the configuration file.
It interacts directly on the configuration file,
and acts as a mediator between Instance Manager and the global configuration file.

"""
import configparser
import os

from recompute import utils

# configuration
CONFIG_FILE = os.path.join(os.environ['HOME'], '.recompute.conf')
# setup logger
logger = utils.get_logger(__name__)


class ConfigManager(object):
  """ConfigManager encapsulates the global config file."""

  def __init__(self, configfile=None):
    """
    Parameters
    ----------
    configfile : str, optional
      Path to the global config file (default None).
      By default, ConfigManager reads from `$HOME/.recompute.conf`.
    """
    # resolve config file
    self.configfile = configfile if configfile else CONFIG_FILE
    # read config from config file
    self.config = self.load() if self.config_exists else None

  def load(self, configfile=None):
    """Load config dictionary from config file.

    Parameters
    ----------
    configfile : str, optional
      Path to the global config file (default None).
      By default, `$HOME/.recompute.conf` is read.

    Returns
    -------
    configparser.ConfigParser
      ConfigParser instance read from `configfile`
    """
    # resolve config file
    configfile = configfile if configfile else self.configfile
    # create parser instance
    config = configparser.ConfigParser()
    # read from file
    config.read(self.configfile)
    return config

  def config_exists(self):
    """Does global config file exist?"""
    return os.path.exists(self.configfile)

  def generate(self, force=False, configfile=None):
    """Generate a sample config file

    Parameters
    ----------
    force : bool, optional
      When set to `True`, permits overwriting global config file (default False)
      Caution adviced. Your `$HOME/.recompute.conf` will be overwritten
    configfile : str, optional
      Path to the global config file (default None).
      By default, sample config is written to `$HOME/.recompute.conf`

    Returns
    -------
    configparser.ConfigParser
      ConfigParser instance read from generate `configfile`
    """
    # if config file exits
    if self.config_exists() and not force and not configfile:
      logger.info('Config File exists; add --force to overwrite')
      return
    # create config
    config = configparser.ConfigParser()
    config.add_section('general')         # add general section
    config.set('general', 'instance', '0')  # default instance
    config.set('general', 'remote_home', 'projects/')  # remote home folder

    # resolve config file
    configfile = configfile if configfile else self.configfile
    # write to config file
    config.write(open(configfile, 'w'))
    logger.info('config written to [{}]'.format(self.configfile))
    logger.info(dict(config['general']))
    # attach to self
    self.config = config
    return config

  def update(self, config):
    """Update global config file with `config`

    Parameters
    ----------
    config : configparser.ConfigParser
      New configuration to be written to global config file
    """
    self.config = config
    # write to config file
    self.config.write(open(self.configfile, 'w'))
    logger.debug('config file updated')

  def get_instances(self):
    """Return a list of instances read from config file

    Returns
    -------
    list
      A list of (SectionProxy) instance sections read from config file
    """
    return [ self.config[section]
        for section in self.config.sections()
        if 'instance ' in section ]

  def get_instance(self, idx=None):
    """
    Parameters
    ----------
    idx : int, optional
      Index of instance to get from config file (default None)

    Returns
    -------
    configparser.SectionProxy
      An instance section from config file

    """
    if idx is None:
      idx = int(self.config['general']['instance'])
    try:
      self.config['instance {}'.format(idx)], 'No such instance in config'
      return self.config['instance {}'.format(idx)]
    except KeyError:
      print('There is no "instance {}" section in config'.format(idx))

  def get_default_instance(self):
    """Return default instance section from config file

    Returns
    -------
    configparser.SectionProxy
      The default instance section from config file
    """
    idx = int(self.config['general']['instance'])
    return self.get_instance(idx)

  def add_instance(self, instance):
    """Add an instance section to config file

    Parameters
    ----------
    instance : instance.Instance
      A new Instance to be added to config file
    """
    # get next index
    idx = len([ sec for sec in self.config.sections()
      if 'instance ' in sec ])
    # add new instance section to config file
    self.config['instance {}'.format(idx)] = {
        'username' : instance.username,
        'host' : instance.host,
        'password' : instance.password
        }
    # update config file
    self.update(self.config)



"""bundle.py

Bundle encapsulates the local repository in the current directory.
Dependencies are resolved and added to a local database (`.recompute/resync.db`).
pypi package requirements are resolved and added to `requirements.txt`.
Inclusion/Exclusion rules are read from the local configuration files `.recompute/include` and `.recompute/exclude`.

"""
import os

from recompute import process
from recompute import cmd
from recompute import utils

# setup logger
logger = utils.get_logger(__name__)

# local configuration files
RSYNC_DB = '.recompute/rsync.db'
REQS = 'requirements.txt'
INCLUDE = '.recompute/include'
EXCLUDE = '.recompute/exclude'


class Bundle(object):
  """Bundle encapsulates local repository in current folder."""

  def __init__(self, name=None):
    """
    Parameters
    ----------
    name : str, optional
      The name (parent) of current directory (default None)
    """
    # get current path
    self.path = os.path.abspath('.')
    # get name of current directory if custom name isn't given
    self.name = name if name else self.path.split('/')[-1]
    # local db
    self.db = RSYNC_DB
    # init include/exclude files
    self.init_include_exclude()
    # update bundle dependencies
    self.update_dependencies()

  def update_dependencies(self):
    """Update dependencies including local files and pypi packages."""
    # get a list of files (local dependencies)
    self.files = list(self.get_local_deps())
    # add INCLUDE; remote EXCLUDE
    self.files = self.inclusion_exclusion()
    logger.info(' '.join(self.files))
    # create a file containing list of dependencies
    self.populate_requirements()
    # get a list of dependencies (python packages)
    self.requirements = self.get_requirements()
    # create a file containting list of local dependencies
    self.populate_local_deps()

  def init_include_exclude(self):
    """Create include/exclude local configuration files."""
    if not os.path.exists(INCLUDE):
      open(INCLUDE, 'w').close()
    if not os.path.exists(EXCLUDE):
      open(EXCLUDE, 'w').close()

  def inclusion_exclusion(self):
    """Read from include/exclude configuration files and update database."""
    # read from .recompute/include
    include = [ f for f in open(INCLUDE).readlines() if f.strip() ]
    # read from .recompute/exclude
    exclude = [ f for f in open(EXCLUDE).readlines() if f.strip() ]
    # update self.files
    self.files = [ f for f in self.files if f not in exclude ]
    self.files.extend(include)
    return list(set(self.files))

  def get_local_deps(self):
    """Resolve local dependencies.

    Run a search for *.py files in current directory.
    """
    for dirpath, dirnames, filenames in os.walk("."):
      for filename in [f for f in filenames if f.endswith(".py")]:
        yield os.path.join(dirpath, filename)

  def populate_local_deps(self):
    """ Write local dependencies to file (local database)."""
    with open(self.db, 'w') as db:
      for filename in self.files:
        logger.info(filename)
        db.write(filename)
        db.write('\n')

  def populate_requirements(self):
    """Resolve pypi package dependencies and populate requirement.txt."""
    # . get a list of pip packages
    # .. write to requirements.txt
    assert process.execute(
        cmd.REDIRECT_STDOUT_NULL.format(command=cmd.PIP_REQS)
        )

  def get_requirements(self):
    """Read from requirements.txt."""
    reqs = [ line.replace('\n', '')
        for line in open(REQS).readlines()
        if line.replace('\n', '').strip()
        ]
    logger.info(reqs)
    return reqs



"""utils.py : A suite of helper functions"""

from prettytable import PrettyTable

import os
import logging
import random

# setup local configuration
LOCAL_CONFIG_DIR = '.recompute'
if not os.path.exists(LOCAL_CONFIG_DIR):
  os.makedirs(LOCAL_CONFIG_DIR)
# redirect log to
LOG = '.recompute/log'
LOG_OVERFLOW = 2000


def get_logger(name, level=logging.INFO):
  """Configure logging mechanism and return logger instance

  Parameters
  ----------
  name : str
    module name (__name__)
  level : int, optional
    Level of logging (default logging.INFO)

  Returns
  -------
  logging.Logger
    A logger instance
  """
  # . if log file doesn't exist
  # .. or if it overflows
  if not os.path.exists(LOG) or len(open(LOG).readlines()) > LOG_OVERFLOW:
    open(LOG, 'w').close()  # create anew
  logging.basicConfig(filename=LOG, filemode='a', level=level)
  return logging.getLogger(name)


# ---- logger ----
logger = get_logger(__name__)


def parse_log(log):
  """Parse log

  Parameters
  ----------
  log : str
    Contents of log file

  Returns
  -------
  str
    Parsed log file
  """
  return log  # TODO : clean up log


def tabulate_processes(processes):
  """Convert list of processes into a Pretty Table

  Parameters
  ----------
  processes : list
    List of processes (name, pid)

  Returns
  -------
  PrettyTable
    A table of processes
  """
  table = PrettyTable()
  table.field_names = [ "Index", "Name", "PID" ]
  # fabricate 0th row
  table.add_row((0, 'all', '*'))
  for idx, (name, pid) in enumerate(processes):
    table.add_row((idx + 1, name, pid))
  return table


def tabulate_instances(instances):
  """Convert a dictionary of instances into a Pretty Table

  Parameters
  ----------
  instances : dict
    Dictionary of instances

  Returns
  -------
  PrettyTable
    A table of instances
  """
  # create pretty table
  table = PrettyTable()
  # add fields
  table.field_names = [ "Machine", "Status", "GPU (MB)", "Disk (MB)" ]
  # add rows
  for instance, values in instances.items():
    table.add_row(values)
  return table


def resolve_relative_path(filename, path):
  """Convert relative path to absolute"""
  return os.path.join(path, filename)


def resolve_absolute_path(filename):
  """Fetch filename from absolute path"""
  return filename.split('/')[-1]


def chain_commands(commands):
  """Chain commands together using `&&` "and" operator

  Parameters
  ----------
  commands : list
    A list of commands to be chained together

  Returns
  -------
  str
    Chained command
  """
  # separate them
  if isinstance(commands, type('42')):
    commands = commands.split('&&')
  # get root command
  root = commands[0]
  # add exec to subsequent commands
  subseq = [ 'exec {}'.format(command) for command in commands[1:] ]
  # chain them
  return root + ' ' + ' && '.join(subseq)


def parse_ps_results(stdout):
  """Parse result of `ps` command

  Parameters
  ----------
  stdout : str
    Output of running `ps` command

  Returns
  -------
  list
    A List of process id's
  """
  # ps returns nothing
  if not stdout.replace('\n', '').strip():
    return []
  # ps returns something
  return [ int(line.split()[0]) for line in stdout.split('\n')
      if line.replace('\n', '').strip()
      ]


def parse_free_results(stdout):
  """Parse results of `free` command

  Parameters
  ----------
  stdout : str
    Output of running `free` command

  Returns
  -------
  int
    Free Disk space
  """
  line = stdout.split('\n')[1]
  assert 'Mem:' in line
  return int(line.split()[3])


def rand_server_port(a=8824, b=8850):
  """Get a random integer between `a` and `b`"""
  return random.randint(a, b)


def rand_client_port(a=8850, b=8890):
  """Get a random integer between `a` and `b`"""
  return random.randint(a, b)



"""cmd.py

Unorganized messy blob of commands.

TODO
----
This definitely requires better organization.
Commands in string format should be replaced with functions with arguments for formatting the command.

"""
# __pipreqs__ resolves pypi package requirements
# and writes it to requirements.txt
PIP_REQS = 'pipreqs . --force'

# __nvidia-smi__ gives a status report on the GPU
# we format the results to get free GPU memory
GPU_FREE_MEMORY = 'nvidia-smi \
    --query-gpu=memory.free \
    --format=csv,nounits,noheader'

# __free__ outputs free disk space available in device
# we format it using `utils.parse_free_results`
DISK_FREE_MEMORY = "free -m"

# __jupyter-notebook__ starts a notebook server in remote device
# `port_num` acts as a handle to keep track of the server
JUPYTER_SERVER = 'jupyter-notebook --no-browser --port={port_num}\
    --NotebookApp.token="" .'

SSH_HEADER = 'sshpass -p {password}'

# we hook up the remote notebook server to a local port via __ssh__
# `client_port_num` and `server_port_num` are required
JUPYTER_CLIENT = SSH_HEADER + ' ' + 'ssh -N -L \
    {client_port_num}:localhost:{server_port_num} \
    {username}@{host}'

# __ps__ results are filtered using __grep__ with a the pattern __re.runner__
# to identify the processes we started in the remote system
PROCESS_LIST_LINUX = 'ps axf | grep re.runner | grep -v grep'

# process list for __mac__
# ...
PROCESS_LIST_OSX = 'ps aux | grep re.runner | grep -v grep'

# filter __ps__ results using __pid__
# ...
PROCESS_PID_LINUX = 'ps -aux | grep {pid}'

# __kill__ process with pid as handle
# ...
KILL_PROCESS = 'kill -9 {}'

# __scp__ copies file from remote device to local device
# one file at a time, fellas!
SCP_FROM_REMOTE = 'scp -r {username}@{host}:{remotepath} {localpath}'

# __scp__ copies file from local device to remote device
# one file at a time, fellas!
SCP_TO_REMOTE = 'scp -r {localpath} {username}@{host}:{remotepath}'

# __rsync__ synchonizes files listed in `.recompute/rsync.db`
# with remote device
RSYNC = 'rsync -a --files-from={deps_file} . \
        {username}@{host}:{remote_dir}'

# execute __cmd__ in remote device via __ssh__
# ...
SSH_EXEC = 'ssh {username}@{host} \'{cmd}\''

# __ssh__ execute in remote device with a pseudo tty terminal
# ...
SSH_EXEC_PSEUDO_TERMINAL = 'ssh -t {username}@{host} \'{cmd}\''

# __nohup__ ensures uninterrupted remote execution
# ...
SSH_EXEC_ASYNC = 'ssh {username}@{host} \'nohup {cmd} > {logfile} \
    2>{logfile} & echo $!\''

# start __ssh__ session
# changed into `remote_dir`
__SSH_INTO_REMOTE_DIR = 'ssh -t {username}@{host} \
            "cd {remote_dir}; exec \\$SHELL --login"'
SSH_INTO_REMOTE_DIR = SSH_HEADER + ' ' + __SSH_INTO_REMOTE_DIR

# with __mkdir__, create directory in remote device
# ...
SSH_MAKE_DIR = 'ssh {username}@{host} mkdir -p {remote_dir}'

# run `exit` in a __ssh__ session
# to test if the instance works
SSH_TEST = 'ssh {username}@{host} \'exit\''

# redirect __stdout__ and __stderr__ to `logfile`
# push process to background using __&__
CMD_LOG_FOOTER_ASYNC = ' > {logfile} 2>{logfile} &'

# redirect __stdout__ and __stderr__ to `logfile`
CMD_LOG_FOOTER = ' > {logfile} 2>{logfile}'

# use __date__ to get "laste modified" time-stamp of a file
# ...
LAST_MODIFIED = 'date -r {filename}'

# __cd__ into a path
# NOTE : do we need this? seems silly!
CD = 'cd {path}'

# set `SIGINT` __trap__
# execute `kill` when SIGINT signal is received
TRAP = 'trap "kill 0" SIGINT'

# run `command` and redirect __stdout__ and __stderr__ to `logfile`
# ...
REDIRECT_STDOUT = '{command} > {logfile} 2>&1'

# run `command` and redirect __stdout__ and __stderr__ to /dev/null
# we basically throw away outputs
REDIRECT_STDOUT_NULL = '{command} > /dev/null 2>&1'

# __wait__ for user interruption
# NOTE : again, do we need this?
WAIT = 'wait'

# execute bash script `runner`
# ...
EXEC_RUNNER = 'bash {runner}'

# set `INT` and `TERM` __trap__
# exit when trap is triggered
TRAP_INT_TERM = 'trap "exit" INT TERM'

# set `EXIT` __trap__
# `kill` process when trap is triggered
TRAP_EXIT = 'trap "kill 0" EXIT'


def make_wget(urls):
  """Download multiple URLs with `wget`

  Parameters
  ----------
  urls : list
    A list of URLs to download from
  """
  return 'wget -c {}'.format(' '.join(urls))


def make_traps():
  """Return a list of traps.

  We place these traps, one in a line, in the `runner` script.
  """
  return [ TRAP_INT_TERM, TRAP_EXIT ]


def kill_procs(pids):
  """ Kill a list of processes using their PIDs

  Parameters
  ----------
  pids : list
    A list of process id's (processes that ought to be killed)
  """
  return ' '.join( ['kill -9'] + [ str(pid) for pid in pids ] )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="recompute",
    version="0.9.10",
    author="Suriyadeepan Ramamoorthy",
    author_email="suriyadeepan.r@gmail.com",
    description="Remote Computation Framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/suriyadeepan/recompute.py",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
      'console_scripts' : [ 're=recompute.recompute:main' ],
      }
)



"""remote.py

Remote class models the remote machine.
It manages the interaction between the remote and the local.
Interaction includes,

* File transfer
* Log management
* Remote Execution
* Environment setup
* Process Management

"""
from __future__ import print_function
import time
import pickle
import os

from recompute import cmd
from recompute import process
from recompute import utils

# setup logger
logger = utils.get_logger(__name__)
# void cache
VOID_CACHE = '.recompute/void'


class Remote(object):
  """Remote models the remote machine"""

  def __init__(self, instance=None, bundle=None, remote_home=None):
    """
    Parameters
    ----------
    instance : instance.Instance, optional
      The instance corresponding to remote device (default None)
    bundle : bundle.Bundle, optional
      Bundle object that encapsulates current working directory (default None)
    remote_home : str, optional
      Home directory of remote device (default None)
    """
    # cache name
    self.CACHE = VOID_CACHE
    cache = None

    if not instance and not bundle:
      # read from cache
      assert os.path.exists(self.CACHE)
      cache = pickle.load(open(self.CACHE, 'rb'))
      remote_home = cache['remote_home']

    self.instance = instance if instance else cache['instance']
    self.bundle = bundle if bundle else cache['bundle']

    # create an SSH Client
    self.client = None

    # projects/ folder in remote machine
    remote_home = remote_home if remote_home else 'projects/'
    self.remote_home = os.path.join(self.get_remote_home_dir(), remote_home)
    # projects/project/ folder in remote machine
    self.remote_dir = os.path.join(self.remote_home, self.bundle.name)
    # projects/project/data/
    self.remote_data = os.path.join(self.remote_dir, 'data/')

    if not cache:
      # make directories in remote machine
      self.make_dirs()

    # build remote log file path
    self.logfile = os.path.join(self.remote_dir,
        '{}.log'.format(self.bundle.name)
      )
    # build local log file path
    self.local_logfile = os.path.join(
        self.bundle.path,
        self.logfile.split('/')[-1]
        )

    # list spawned processes
    self.processes = [] if not cache else cache['processes']
    # cache void
    self.cache_()

  def get_remote_home_dir(self):
    """Get $HOME directory path from remote system"""
    pid, output = process.remote_execute('pwd', self.instance)
    return output.replace('\n', '').strip()

  def cache_(self, name=None):
    """Cache attributes of self (Remote).

    Parameters
    ----------
    name : str, optional
      filename of local cache (default None)
    """
    # get cache name
    name = self.CACHE if not name else name
    # create a minimal dict of self
    void_as_dict = {
        'instance' : self.instance,
        'bundle': self.bundle,
        'remote_home' : self.remote_home,
        'processes' : self.processes
        }
    # dump dictionary
    pickle.dump(void_as_dict, open(self.CACHE, 'wb'))

  def make_mkcmd(self, dir_=None):
    """Make mkdir command

    Parameters
    ----------
    dir_ : str
      Directory to create

    Returns
    -------
    str
      Command to make directory `dir_` in remote machine
    """
    # resolve directory to make
    dir_ = dir_ if dir_ else self.remote_dir
    # create directory in remote machine
    _header = cmd.SSH_HEADER.format(password=self.instance.password)
    _body = cmd.SSH_MAKE_DIR.format(
        username=self.instance.username,
        host=self.instance.host,
        remote_dir=dir_
        )
    # join
    return ' '.join([_header, _body])

  def make_rsync_cmd(self):
    """Make rsync command

    Returns
    -------
    str
      rsync-based command that copies local files to remote
    """
    _header = cmd.SSH_HEADER.format(password=self.instance.password)
    _body = cmd.RSYNC.format(
            deps_file=self.bundle.db,
            username=self.instance.username,
            host=self.instance.host,
            remote_dir=self.remote_dir
            )

    return ' '.join([_header, _body])

  def make_dirs(self):
    """Make necessary directories in remote machine"""
    # execute make directory command from local machine
    mkcmd = self.make_mkcmd()
    logger.info(mkcmd)
    assert process.execute(mkcmd)
    # make data/ directory
    remote_data_ = self.make_mkcmd(self.remote_data)
    logger.info(remote_data_)
    assert process.execute(remote_data_)

  def rsync(self, update=False):
    """Rsync files between local and remote systems

    Parameters
    ----------
    update : bool, optional
      When set to `True` updates dependencies in local database (default False)

    Returns
    -------
    tuple
      (pid, output) Process id and output of execution
    """
    if update:  # update bundle
      self.bundle.update_dependencies()

    # execute rsync
    rsync_cmd = self.make_rsync_cmd()
    logger.info(rsync_cmd)
    return process.execute(rsync_cmd)

  def async_execute(self, commands, logfile=None, name='runner'):
    return self.execute(commands, run_async=True, log=True, logfile=logfile, name=name)

  def execute(self, commands, run_async=False, log=True, logfile=None, name='runner'):
    """Execute `cmdstr` in remote device given by `instance`

    Parameters
    ----------
    commands : list
      List of commands to be executed sequentially
    run_async : bool, optional
      When set to `True` runs commands asynchronously
      When set to `False` runs commands synchronously (default False)
    log : bool, optional
      Enables or disables logging output of execution (default True)
    logfile : str, optional
      Log file to redirect output of execution to (default None)
    name : str, optional
      Name of process (default 'runner')

    Returns
    -------
    tuple
      (pid, output) Process id and STDOUT of execution
    """
    # resolve log file
    logfile = logfile if logfile else self.logfile
    # create runner
    runner = process.create_runner(self.remote_dir, commands, logfile, run_async=run_async)
    # push runner to remote
    self.copy_file_to_remote(
        os.path.join(self.bundle.path, runner),  # abs path of current dir
        self.remote_dir
        )
    # get absolute path of runner
    runner_abs_path = os.path.join(self.remote_dir, runner)
    # execute runner in remote machine
    exec_fn = process.remote_async_execute if run_async else process.remote_execute
    pid, output = exec_fn(cmd.EXEC_RUNNER.format(runner=runner_abs_path), instance=self.instance)

    # add pid to processes
    self.processes.append((name, pid))
    self.cache_()
    return pid, output

  def execute_command(self, cmdstr, run_async=False,
      log=False, logfile=None, bypass_subprocess=True):
    """Execute `cmdstr` in remote device

    Parameters
    ----------
    cmdstr : str
      Command to be executed
    run_async : bool, optional
      When set to `True` runs commands asynchronously
      When set to `False` runs commands synchronously (default False)
    log : bool, optional
      Enables or disables logging output of execution (default True)
    logfile : str, optional
      Log file to redirect output of execution to (default None)

    bypass_subprocess : bool, optional
      When set to `True`, `os.system` is used for execution, (None, None) is returned
      When set to `False`, subprocess module is used for execution (default True)

    Returns
    -------
    tuple
      (pid, output) Process id and STDOUT of execution
    """
    logger.info(cmdstr)
    # if we are running async
    if run_async:
      return process.remote_async_execute(cmdstr, self.instance, logfile)
    # --- sync execution --- #
    # execute in remote machine
    return process.remote_execute(cmdstr, self.instance,
        bypass_subprocess=bypass_subprocess)

  def async_execute_command(self, cmdstr, logfile=None):
    """Execute `cmdstr` in remote device, asynchronously

    Parameters
    ----------
    cmdstr : str
      Command to be executed
    logfile : str, optional
      Log file to redirect output to (default None)

    Returns
    -------
    tuple
      (pid, output) Process id and STDOUT of execution
    """
    return self.execute_command(cmdstr, run_async=True, log=True, logfile=logfile)

  def copy_file_to_remote(self, localpath, remotepath=None):
    """Copy file to remote machine

    Parameters
    ----------
    localpath : str
      Path to local file to be copied to remote machine
    remotepath : str, optional
      Path in remote machine where local file should be copied to
    """
    # default remote path
    remotepath = remotepath if remotepath else self.remote_data
    # build copy cmd
    _header = cmd.SSH_HEADER.format(password=self.instance.password)
    _body = cmd.SCP_TO_REMOTE.format(
        username=self.instance.username,
        host=self.instance.host,
        remotepath=remotepath,
        localpath=localpath
        )
    copy_cmd = ' '.join([_header, _body])
    # local execute scp
    process.execute(copy_cmd)

  def get_file_from_remote(self, remotepath, localpath=None):
    """Copy file to local machine

    Parameters
    ----------
    remotepath : str
      Path to remote file to be copied to local machine
    remotepath : str, optional
      Path in local machine where remote file should be copied to
    """
    # default local path
    localpath = localpath if localpath else self.bundle.path
    # build copy cmd
    _header = cmd.SSH_HEADER.format(password=self.instance.password)
    _body = cmd.SCP_FROM_REMOTE.format(
        username=self.instance.username,
        host=self.instance.host,
        remotepath=remotepath,
        localpath=localpath
        )
    copy_cmd = ' '.join([_header, _body])
    # execute scp command
    process.execute(copy_cmd)

  def get_remote_log(self, keyword=None):
    """Copy log file in remote system to local machine

    Parameters
    ----------
    keyword : str, optional
      A keyword to filter out log file (default None)

    Returns
    -------
    str
      Contents of log file in remote machine
    """
    # copy to local
    self.get_file_from_remote(self.logfile, self.bundle.path)  # '.'
    return self.get_local_log(keyword)

  def get_local_log(self, keyword=None):
    """Read local log file

    Parameters
    ----------
    keyword : str, optional
      A keyword to filter out log file (default None)

    Returns
    -------
    str
      Contents of log file in local machine
    """
    # check if local log file exists
    if not os.path.exists(self.local_logfile):
      return ''

    # read from log file
    log = open(os.path.join(self.bundle.path, self.logfile.split('/')[-1])).read()

    if keyword:    # if keyword is given
      log = '\n'.join([ line for line in log.split('\n') if keyword in line])

    logger.info('\n{}'.format(log))
    return log

  def loop_get_remote_log(self, delay, keyword=None):
    """ Fetch log file in a loop.

    Get log from remote machine every `delay` seconds.
    Print only the difference between local and remote log.
    End loop when "EOF" is seen in log file.

    Parameters
    ----------
    delay : int
      Number of seconds to wait till next fetch of log
    keyword : str, optional
      A keyword to filter out log file (default None)
    """
    try:
      while True:  # tis a loop, my liege.
        # . get local log
        local_log = self.get_local_log()
        # .. get remote log
        remote_log = self.get_remote_log()
        # ... compare
        diff = remote_log.replace(local_log, '')

        # print the diff
        if diff.strip():  # if there is a difference
          logger.info('\n{}'.format(diff))
          print(diff, end='')

        if 'EOF' in diff:  # has the execution ended?
          break
        # and now we wait
        time.sleep(delay)
    except KeyboardInterrupt:
      logger.info('You did this! You did this to us!!')

  def install_deps(self, update=False):
    """Install dependencies in remote system

    Parameters
    ----------
    update : bool, optional
      when set to `True`, dependencies are updated before "pip install"
      (default False)
    """
    if update:  # update bundle
      self.bundle.update_dependencies()
    # install
    self.install(self.bundle.get_requirements())

  def install(self, packages):
    """Install pypi `packages` in remote system

    Parameters
    ----------
    package : list
      List of pypi packages read from "requirements.txt"
    """
    if len(packages) == 0:  # check if packages list is empty
      logger.info('No pypi packages required for execution')
      return
    # make pip-install command
    pip_install_cmd = 'python3 -m pip install --user {}'.format(
        ' '.join(packages)
        )
    logger.info('\tInstall dependencies\n\t{}'.format(pip_install_cmd))
    # remote execute cmd
    logger.info('\n\t{}'.format(self.execute_command(pip_install_cmd)))

  def _header_cd(self, dir_=None):
    """Create "change directory" header

    Parameters
    ----------
    dir_ : directory to change to

    Returns
    -------
    str
      Command to change directory
    """
    # resolve directory to change to
    dir_ = dir_ if dir_ else self.remote_dir
    return 'cd {}'.format(dir_)

  def _header_cd_data(self):
    """Change Directory to data/ header

    Returns
    -------
    str
      Command to change to "data/" directory in remote machine
    """
    return 'cd {}'.format(self.remote_data)

  def download(self, urls, change_to=None, run_async=False):
    """Download from web to remote machine's "data/" directory

    Parameters
    ----------
    urls : list
      List of URLs to download from
    change_to : str, optional
      Directory where downloaded files go (default None)
    run_async : bool, optional
      Download asynchronously (default False)
    """
    # resolve cd directory
    change_to = change_to if change_to else self.remote_data
    # build cmd header
    _cmd_header = self._header_cd(change_to)
    # build cmd body
    _cmd_body = cmd.make_wget(urls)
    # build cmd footer
    data_logfile = os.path.join(change_to, 'data.log')
    _cmd_footer = cmd.CMD_LOG_FOOTER.format(logfile=data_logfile)
    # join
    cmd_str = ' && '.join([ _cmd_header, _cmd_body ])

    if run_async:  # run (download) async
      self.async_remote_exec(cmd_str, logfile=data_logfile)

    # run (download) sync
    logger.info(self.execute_command(cmd_str + _cmd_footer))

  def get_session(self):
    """Create an ssh session"""
    os.system(
        cmd.SSH_INTO_REMOTE_DIR.format(
          username=self.instance.username,
          password=self.instance.password,
          host=self.instance.host,
          remote_dir=self.remote_dir
          )
        )

  def start_notebook(self, run_async=False, name='jupyter:{}'):
    """Create and connect to remote notebook server

    Parameters
    ----------
    run_async : bool, optional
      (currently unsupported) Enables asynchronous execution of local notebook (default False)
    name : str, optional
      Name of remote notebook server process (default jupyter:{})
    """
    # TODO : what if notebook isn't installed ? Most unlikely uninstalled!
    # install jupyter notebook
    # self.install(['jupyter'])
    # choose a port number
    server_port_num = utils.rand_server_port()
    # set name=name.format(server_port_num))
    notebook_server_name = name.format(server_port_num)
    # start jupyter server
    commands = [ cmd.JUPYTER_SERVER.format(port_num=server_port_num) ]
    notebook_server_pid, output = self.async_execute(commands,
        logfile='/dev/null',        # don't need no log
        name=notebook_server_name)  # name it, so we can track it

    logger.info('\tStarted notebook server in remote machine :{}'.format(
      server_port_num))
    print('{username}@{host}:{port}'.format(
      username=self.instance.username,
      host=self.instance.host,
      port=server_port_num))

    # . choose a client port number
    # .. build notebook client command
    client_port_num = utils.rand_client_port()
    cmd_local = cmd.JUPYTER_CLIENT.format(
        username=self.instance.username,
        password=self.instance.password,
        host=self.instance.host,
        client_port_num=client_port_num,
        server_port_num=server_port_num
        )

    logger.info('Starting local notebook ')
    logger.info('\thttp://localhost:{}/tree'.format(client_port_num))
    print('http://localhost:{port}/tree'.format(port=client_port_num))

    if run_async:
      print('Async Execution not implemented yet!')  # TODO : add issue handle here
      # pid, output = process.async_execute(cmd_local)
      # return

    # connect to notebook server
    pid, output = process.execute(cmd_local)

    if not pid and not output:  # keyboard interrupt
      logger.info('YOU quit jupyter notebook')
      # ----- kill server -----
      # . list processes
      procs = self.list_processes()
      # .. find jupyter notebook process
      name, pid = procs[-1]  # pretty sure it's the last added process
      # ... make sure name and pid check out
      assert pid == notebook_server_pid
      assert name == notebook_server_name
      # .... get handle
      notebook_server_idx = len(procs) - 1
      # ..... kill it
    self.kill(notebook_server_idx)  # if its 0, no problem!

  def list_processes(self, force=False):
    """
    Parameters
    ----------
    force : bool, optional
      When set to `True`, queries remote machine to get running processes
      When set to `False`, returns existing processes (default False)

    Returns
    -------
    list
      A list of processes active in remote machine [ (name, pid) ... ]

    """
    if force:
      # get list of processes
      command = ' '.join([
        cmd.SSH_HEADER.format(password=self.instance.password),
        cmd.SSH_EXEC.format(username=self.instance.username, host=self.instance.host,
          cmd=cmd.PROCESS_LIST_LINUX.format(username=self.instance.username)
          )
        ])
      pid, output = process.execute(command)
      pids = utils.parse_ps_results(output)
      known_pids = [ proc[-1] for proc in self.processes ]
      self.processes = [ (name, pid) for name, pid in self.processes
          if pid in pids ]
      self.processes.extend([ ('zombie/spawn', pid) for pid in pids
        if pid not in known_pids ])
      # update cache
      self.cache_()

    logger.info(self.processes)
    return self.processes

  def kill(self, idx, force=False):
    """Kill process by index

    Parameters
    ----------
    idx : int
      Index of process to kill
    force : bool, optional
      When set to `True`, queries remote machine to get running processes
      When set to `False`, considers existing processes (default False)
    """
    processes = self.list_processes(force=force)
    if len(processes) > 0:
      procs_to_kill = processes if idx == 0 else [processes[idx - 1]]
      if len(procs_to_kill) > 0:
        process.kill_remote_process(
            [ p[-1] for p in procs_to_kill ],  # separate pid
            instance=self.instance)
